using System;
using System.Diagnostics;
using System.Drawing;
using System.Windows.Forms;

namespace iclect.Forms
{
	/// <summary>
	/// Summary description for FilmPanelManager.
	/// </summary>
	internal class FilmPanels
	{
		Film			_film;
		FilmPanel []	_panels;
		public FilmPanels(Film film)
		{
			_film		= film;
			_panels		= new FilmPanel[2];
			ThumbSize	= 96;
		}
		public void SetPanels(FilmPanel panel0, FilmPanel panel1)
		{
			panel0.ID	= 0;
			panel1.ID	= 1;
			_panels[0]	= panel0;
			_panels[1]	= panel1;

			foreach (FilmPanel panel in _panels)
			{
				panel.Paint		+= new PaintEventHandler(OnPaintPanel);
				panel.MouseDown	+= new MouseEventHandler(OnPanelMouseDown);
				panel.MouseMove	+= new MouseEventHandler(OnPanelMouseMove);
			}
		}

		Orientation _orientation;
		int _xf;
		int _yf;
		public Orientation Orientation
		{
			get { return _orientation; }
			set
			{
				_orientation = value;
				if (_orientation == Orientation.Horizontal)
				{
					_xf					= 1;
					_yf					= 0;
				}
				else
				{
					_xf					= 0;
					_yf					= 1;
				}
				_film.PerformLayout();
			}
		}

		const int	_thumbBorder	= 2;
		const int	_thumbMargin	= 1;
		const int	_punchSize		= 4;
		const int	_punchBorder	= 2;
		const int	_punchSkip		= 8;

		int			_thumbSize;
		int			_thumbTotal;	// _thumbSize  + 2 * _thumbBorder
		int			_thumbSkip;		// _thumbTotal + 2 * _thumbMargin
		int			_filmWidth;		// _thumbTotal + 2 * _punchSkip

		int			_maxPanelHeight;
		int			_maxThumbsPerPanel;

		public int ThumbSize
		{
			get { return _thumbSize; }
			set
			{
				if (_thumbSize != value)
				{
					_thumbSize			= value;
					_thumbTotal			= _thumbSize  + 2 * _thumbBorder;
					_thumbSkip			= _thumbTotal + 2 * _thumbMargin;
					_filmWidth			= _thumbTotal + 2 * _punchSkip;

					Rectangle vScreen	= SystemInformation.VirtualScreen;
					int minHeight		= Math.Max(vScreen.Width, vScreen.Height);
					int skipProd		= _thumbSkip * _punchSkip;
					int countSkips		= (minHeight + skipProd - 1) / skipProd;
					_maxPanelHeight		= countSkips * skipProd;
					_maxThumbsPerPanel	= _maxPanelHeight / _thumbSkip;
					_film.PerformLayout();
				}
			}
		}
		public int ThumbSkip
		{
			get { return _thumbSkip; }
		}
		public int FilmWidth
		{
			get { return _filmWidth; }
		}
		public int FilmHeight
		{
			get { return _film.ImageCount * _thumbSkip; }
		}
		public int ViewHeight
		{
			get
			{
				if (Orientation == Orientation.Vertical)
					return _film.Height;
				else
					return _film.Width;
			}
		}
		public int FirstVisibleIndex
		{
			get { return (_scrollValue + _thumbSkip - 1) / _thumbSkip; }
		}
		public int LastVisibleIndex
		{
			get { return (_scrollValue + ViewHeight - _thumbSkip) / _thumbSkip; }
		}
		public int ViewCount
		{
			get { return ViewHeight / _thumbSkip; }
		}

		internal struct ThumbID
		{
			public int PanelID;
			public int PanelIndex;
			public int ThumbIndex;
		}
		internal ThumbID GetThumbID(int index)
		{
			ThumbID id		= new ThumbID();
			id.PanelID		= index / _maxThumbsPerPanel;
			id.ThumbIndex	= index % _maxThumbsPerPanel;
			id.PanelIndex	= id.PanelID % 2;
			return id;
		}

		public void Invalidate()
		{
			foreach (FilmPanel panel in _panels)
                panel.Invalidate();
		}
		public void Invalidate(int index)
		{
			ThumbID id		= GetThumbID(index);
			FilmPanel panel	= _panels[id.PanelIndex];
			if (panel.ID == id.PanelID)
			{
				Rectangle rect = GetThumbRectangle(id.ThumbIndex);
				panel.Invalidate(rect);
			}
		}
		public void InvalidateFocusCues()
		{
			Rectangle rectLeft;
			Rectangle rectRight;
			if (Orientation == Orientation.Vertical)
			{
				rectLeft	= new Rectangle(0, 0, _punchSkip, _maxPanelHeight);
				rectRight	= new Rectangle(_punchSkip + _thumbTotal, 0, _punchSkip, _maxPanelHeight);
			}
			else
			{
				rectLeft	= new Rectangle(0, 0, _maxPanelHeight, _punchSkip);
				rectRight	= new Rectangle(0, _punchSkip + _thumbTotal, _maxPanelHeight, _punchSkip);
			}
			foreach (FilmPanel panel in _panels)
			{
				panel.Invalidate(rectLeft);
				panel.Invalidate(rectRight);
			}
		}
		public void Layout()
		{
			int view				= ViewHeight;
			int firstIndex			= _scrollValue / _thumbSkip;
			ThumbID firstID			= GetThumbID(firstIndex);
			int topPosition			= firstID.PanelID * _maxPanelHeight;
			int panelTop			= -(_scrollValue - topPosition);
			int panelHeight			= Math.Min(FilmHeight - topPosition, _maxPanelHeight);
			FilmPanel firstPanel	= _panels[firstID.PanelIndex];
			firstPanel.ID			= firstID.PanelID;

			if (Orientation == Orientation.Vertical)
			{
				firstPanel.Top		= panelTop;
				firstPanel.Width	= _filmWidth;
				firstPanel.Height	= panelHeight;
			}
			else
			{
				firstPanel.Left		= panelTop;
				firstPanel.Height	= _filmWidth;
				firstPanel.Width	= panelHeight;
			}

			FilmPanel otherPanel	= _panels[1 - firstID.PanelIndex];
			int lastIndex			= (_scrollValue + view - 1) / _thumbSkip;
			ThumbID lastID			= GetThumbID(lastIndex);
			
			if (lastID.PanelID == firstID.PanelID)
			{
				otherPanel.Visible		= false;
			}
			else
			{
				topPosition				+= _maxPanelHeight;
				panelTop				= -(_scrollValue - topPosition);
				panelHeight				= Math.Min(FilmHeight - topPosition, _maxPanelHeight);
				otherPanel.ID			= lastID.PanelID;

				if (Orientation == Orientation.Vertical)
				{
					otherPanel.Top		= panelTop;
					otherPanel.Width	= _filmWidth;
					otherPanel.Height	= panelHeight;
				}
				else
				{
					otherPanel.Left		= panelTop;
					otherPanel.Height	= _filmWidth;
					otherPanel.Width	= panelHeight;
				}

				otherPanel.Visible		= true;
			}

			firstPanel.Visible		= true;
		}
		int _scrollValue = 0;
		public int ScrollValue
		{
			get { return _scrollValue; }
			set
			{
				if (value != _scrollValue)
				{
					int diff = Math.Abs(value - _scrollValue);
					_scrollValue = value;
					if (diff > _thumbSkip / 2)
						_film.Invalidate(true);
					_film.PerformLayout();
				}
			}
		}

		#region Drawing
		Rectangle GetThumbRectangle(int index)
		{
			if (Orientation == Orientation.Horizontal)
				return new Rectangle((index * _thumbSkip) + _thumbMargin, _punchSkip, _thumbTotal, _thumbTotal);
			else
				return new Rectangle(_punchSkip, (index * _thumbSkip) + _thumbMargin, _thumbTotal, _thumbTotal);
		}
		public void CenterThumb(int index)
		{
			int view		= ViewHeight;
			int pos			= index * _thumbSkip + (_thumbSkip - view) / 2;
			int max			= FilmHeight - view;

			if (pos > max)
				pos = max;
			if (pos < 0)
				pos = 0;

            _film.ScrollValue = pos;
		}
		public void ShowThumb(int index)
		{
			int firstIndex	= (_scrollValue + _thumbSkip - 1) / _thumbSkip;
			int lastIndex	= (_scrollValue + ViewHeight - _thumbSkip) / _thumbSkip;
			if (index < firstIndex)
			{
				_film.ScrollValue -= (firstIndex - index) * _thumbSkip;
			}
			else if (index > lastIndex)
			{
				_film.ScrollValue += (index - lastIndex) * _thumbSkip;
			}
		}
		public void OnPaintPanel(object sender, System.Windows.Forms.PaintEventArgs e)
		{
			if (_film.ImageCount == 0)
				return;

			FilmPanel panel = sender as FilmPanel;
			if (panel == null)
				return;

			int xskip	= _punchSkip * _xf;
			int yskip	= _punchSkip * _yf;

			Rectangle rect1 = new Rectangle(_punchBorder, _punchBorder, _punchSize, _punchSize);
			Rectangle rect2 = rect1;
			rect2.Offset(_yf * (_thumbTotal + _punchSkip), _xf * (_thumbTotal + _punchSkip));

			Graphics g = e.Graphics;
			Brush brush;
			if (_film.Focused)
				brush = Brushes.White;
			else
				brush = Brushes.LightGray;

			int punchHoles = _maxPanelHeight / _punchSkip;
			for (int i = 0; i < punchHoles; i++)
			{
				g.FillRectangle(brush, rect1);
				g.FillRectangle(brush, rect2);
				rect1.Offset(xskip, yskip);
				rect2.Offset(xskip, yskip);
			}

			int firstIndex	= panel.ID * _maxThumbsPerPanel;
			int lastIndex	= Math.Min(firstIndex + _maxThumbsPerPanel, _film.ImageCount);

			for (int index = firstIndex; index < lastIndex; index++)
			{
				Rectangle thumbRect = GetThumbRectangle(index % _maxThumbsPerPanel);
				if (thumbRect.IntersectsWith(e.ClipRectangle))
				{
					Image img = _film.ImageFolder.ImageFiles[index].GetThumbImage();
					if (img != null)
					{
						int x = thumbRect.Left + (thumbRect.Width - img.Width) / 2;
						int y = thumbRect.Top + (thumbRect.Height - img.Height) / 2;
						g.DrawImage(img, x, y);
					}
					DrawStatusRect(g, index);
				}
			}
		}
		private void DrawStatusRect(Graphics g, int index)
		{
			bool isSelected	= _film.ImageFolder.ImageFiles[index].Selected;

			Color rectColor;
			if (index == _film.FocusedThumbIndex)
			{
				if (isSelected)
					rectColor = _film.SelectedFocusColor;
				else
					rectColor = _film.FocusColor;
			}
			else
			{
				if (isSelected)
					rectColor = _film.SelectionColor;
				else
					rectColor = _film.FilmColor;
			}

			Rectangle thumbRect	= GetThumbRectangle(index % _maxThumbsPerPanel);
			int halfBorder		= _thumbBorder / 2;
			thumbRect.Width		-= _thumbBorder;
			thumbRect.Height	-= _thumbBorder;
			thumbRect.X			+= halfBorder;
			thumbRect.Y			+= halfBorder;
			g.DrawRectangle(new Pen(rectColor, _thumbBorder), thumbRect);
		}
		#endregion
		#region Mouse events
		public void OnPanelMouseDown(object sender, System.Windows.Forms.MouseEventArgs e)
		{
			if (e.Button == MouseButtons.Left)
			{
				FilmPanel panel = sender as FilmPanel;
				if (panel == null)
					return;

				int index;
				if (Orientation == Orientation.Horizontal)
					index = e.X / _thumbSkip;
				else
					index = e.Y / _thumbSkip;
				index += (panel.ID * _maxThumbsPerPanel);

				_film.FocusedThumbIndex = index;
				_film.Focus();
			}
		}
		int _toolTipIndex = -1;
		public void OnPanelMouseMove(object sender, System.Windows.Forms.MouseEventArgs e)
		{
			FilmPanel panel = sender as FilmPanel;
			if (panel == null)
				return;

			int index;
			if (Orientation == Orientation.Horizontal)
				index = e.X / _thumbSkip;
			else
				index = e.Y / _thumbSkip;
			index += (panel.ID * _maxThumbsPerPanel);

			if ((index != _toolTipIndex)
				&& (0 <= index)
				&& (index < _film.ImageCount))
			{
				_toolTipIndex		= index;
				_film.SetToolTip(panel, index);
			}
		}
		#endregion
	}
}
