using System;
using System.Diagnostics;
using System.Drawing;
using System.Drawing.Imaging;
using System.Globalization;
using System.IO;
using System.Text.RegularExpressions;

using iclect.ImageManager.Cache;
using iclect.ImageManager.Thumbnails;

namespace iclect.ImageManager
{
	/// <summary>
	/// Information about an image file.
	/// </summary>
	[Serializable]
	public class ImageFile : IDisposable
	{
		internal ImageFile(ImageFolder imageFolder, FileInfo fileInfo)
		{
			_folder		= imageFolder;
			_fileInfo	= fileInfo;
		}
		~ImageFile()
		{
			Dispose(false);
		}
		public void Dispose()
		{
			Dispose(true);
			GC.SuppressFinalize(this);
		}
		protected void Dispose(bool disposing)
		{
			if (disposing)
			{
				if (_thumb != null)
				{
					_thumb.Dispose();
					_thumb = null;
				}
				if (_high != null)
				{
					_high.Dispose();
					_high = null;
				}
			}
		}
		public DateTime LastWriteTime
		{
			get { return _fileInfo.LastWriteTime; }
		}
		public DateTime CreationTime
		{
			get { return _fileInfo.CreationTime; }
		}
		public string Name
		{
			get { return _fileInfo.Name; }
		}
		public string FullName
		{
			get { return _fileInfo.FullName; }
		}
		public ImageFolder Folder
		{
			get { return _folder; }
		}
		public Exception LoadException
		{
			get { return _loadException; }
		}
		public bool Selected
		{
			get { return _selected; }
			set { _selected = value; }
		}
		public Image GetThumbImage()
		{
			if (_thumb != null)
				return _thumb;

			lock (_getThumbLock)
			{
				if (_thumb != null)
					return _thumb;

				LoadThumb();
				return _thumb;
			}
		}
		internal Image GetHighImage()
		{
			if (_high != null)
				return _high;

			if ((_loadException != null)
				&& (_fileInfo.LastWriteTime == File.GetLastWriteTime(_fileInfo.FullName)))
				return null;

			lock (_getHighLock)
			{
				if (_high != null)
					return _high;

				if (!LoadHigh())
					return null;

				return _high;
			}
		}
		private void LoadThumb()
		{
			if (_thumb != null)
			{
				Image thumb = _thumb;
				_thumb = null;
				thumb.Dispose();
			}
			using (ImageReference iref = GetImageReference())
			{
				const int thumbSize	= 96;
				Image full			= iref.GetImage();
				if (full != null)
				{
					int width		= thumbSize;
					int height		= thumbSize;

					if (full.Width * full.VerticalResolution > full.Height * full.HorizontalResolution)
						height	= (int)((full.Height * full.HorizontalResolution * thumbSize) / (full.VerticalResolution * full.Width));
					else
						width	= (int)((full.Width * full.VerticalResolution * thumbSize) / (full.HorizontalResolution * full.Height));

					_thumb			= _thumbnailer.CreateThumbnail(full, width, height);
					Debug.WriteLine(string.Format("Created thumb for '{0}'.", _fileInfo.FullName));
				}
				else
				{
					_thumb			= _thumbnailer.CreateDummyImage(thumbSize, thumbSize);
					Debug.WriteLine(string.Format("Created dummy for '{0}'.", _fileInfo.FullName));
				}			
			}
		}
		public void SaveThumb(Stream stream, long quality)
		{
			if (_thumb == null)
				return;

			EncoderParameters eps	= new EncoderParameters(1);
			eps.Param[0]			= new EncoderParameter(Encoder.Quality, quality);

			_thumb.Save(stream, _jpegEncoder, eps);
		}

		private static ImageCodecInfo GetEncoderInfo(String mimeType)
		{
			ImageCodecInfo[] encoders = ImageCodecInfo.GetImageEncoders();
			foreach (ImageCodecInfo encoder in encoders)
			{
				if (encoder.MimeType == mimeType)
					return encoder;
			}
			return null;
		}
		static ImageCodecInfo	_jpegEncoder = GetEncoderInfo("image/jpeg");

		public int Width
		{
			get { return _width; }
		}
		public int Height
		{
			get { return _height; }
		}
		public DateTime RecordingTime
		{
			get { return _recordingTime; }
		}

		private bool LoadHigh()
		{
			DisposeHigh();

			try
			{
				using (Stream file = File.OpenRead(_fileInfo.FullName))
				{
					if (file.Length > int.MaxValue)
						throw new FileLoadException("File too large");

					int length		= (int)file.Length;
					byte [] data	= new byte[length];
					file.Read(data, 0, length);
					_highStream		= new MemoryStream(data);
				}

				_high			= Image.FromStream(_highStream, true);
				_width			= _high.Width;
				_height			= _high.Height;
			}
			catch (Exception ex)
			{
				DisposeHigh();
				Debug.WriteLine(ex.Message);
				_loadException = ex;
				return false;
			}

			try
			{
				PropertyItem [] pis = _high.PropertyItems;
				foreach (PropertyItem pi in pis)
				{
					switch (pi.Id)
					{
						case 0x9003:
							_recordingTime = PropParser.ParseDateTime(pi.Value);
							break;
					}
				}
			}
			catch (Exception ex)
			{
				Debug.WriteLine(ex.Message);
			}
			return true;
		}
		private void DisposeHigh()
		{
			Image high = _high;
			_high = null;
			if (high != null)
				high.Dispose();

			Stream highStream = _highStream;
			_highStream = null;
			if (highStream != null)
				highStream.Close();

			_loadException	= null;
		}
		public ImageReference GetImageReference()
		{
			return new ImageReference(this);
		}
		internal void AddRef()
		{
			System.Threading.Interlocked.Increment(ref _referenceCount);
		}
		internal void Release()
		{
			long val = System.Threading.Interlocked.Decrement(ref _referenceCount);
			if (val == 0)
			{
				lock (_getHighLock)
				{
					if (!_folder._highCache.KeepInCache(this))
						DisposeHigh();
				}
			}
		}

		private class PropParser
		{
			static Regex rexDateTime = new Regex(@"(\d{4}):(\d{2}):(\d{2})\s+(\d{2}):(\d{2}):(\d{2})", RegexOptions.Compiled);
			public static string ParseString(byte [] value)
			{
				return System.Text.Encoding.ASCII.GetString(value);
			}
			public static DateTime ParseDateTime(byte [] value)
			{
				string text = ParseString(value);
				Match m = rexDateTime.Match(text);
				return new DateTime(
					int.Parse(m.Groups[1].Value),
					int.Parse(m.Groups[2].Value),
					int.Parse(m.Groups[3].Value),
					int.Parse(m.Groups[4].Value),
					int.Parse(m.Groups[5].Value),
					int.Parse(m.Groups[6].Value)
					);
			}
		}

		[NonSerialized]	internal ImageFolder	_folder;		// The folder that contains the image file.
		/**/			FileInfo				_fileInfo;		// File information
		
		// Additional image information (from within or from the image)
		/**/			int						_width;
		/**/			int						_height;
		/**/			DateTime				_recordingTime;

		// Status information
		/**/			Exception				_loadException;
		[NonSerialized]	bool					_selected;

		// thumbnail image
		[NonSerialized]	internal Image			_thumb;

		// high resolution image
		[NonSerialized]	internal Image			_high;
		[NonSerialized]	Stream					_highStream;
		[NonSerialized]	internal long			_referenceCount;

		/**/			object					_getThumbLock = new object();
		/**/			object					_getHighLock = new object();

		[NonSerialized]	static Thumbnailer		_thumbnailer = new GdiPlusThumbnailer();
	}
}
