using System;
using System.Collections;
using System.Diagnostics;
using System.Drawing;
using System.Drawing.Imaging;
using System.IO;
using System.Runtime.Serialization;
using System.Runtime.Serialization.Formatters.Binary;

using iclect.ImageManager.Cache;

namespace iclect.ImageManager
{
	/// <summary>
	/// Represents a folder containing images.
	/// </summary>
	public class ImageFolder : IDisposable
	{
		public ImageFolder(string path)
		{
			_directory			= new DirectoryInfo(path);
			if (!_directory.Exists)
				throw new FileNotFoundException("The image folder could not be found", path);

			_thumbQuality		= 100L;
			_thumbLoader		= new ThumbLoader(this);
			_highCache			= new HighCache();
			_thumbFilePath		= Path.Combine(_directory.FullName, "iclect.thumbs");
			_imageFiles			= LoadImageFiles(_thumbFilePath);

			Refresh();
		}
		~ImageFolder()
		{
			Dispose(false);
		}
		public void Dispose()
		{
			Dispose(true);
			GC.SuppressFinalize(this);
		}
		protected void Dispose(bool disposing)
		{
			if (disposing)
			{
				_thumbLoader.Stop();
				if (_imageFiles != null)
				{
					foreach (ImageFile imageFile in _imageFiles)
						imageFile.Dispose();
					_imageFiles = null;
				}
			}
		}
		private void Refresh()
		{
			_thumbLoader.Stop();
			ImageFile [] left	= LoadFileList(_directory);
			ImageFile [] right	= _imageFiles;
			ImageFile [] merged	= Merge(left, right);
			_imageFiles			= merged;
			_thumbLoader.Start();
		}
		public ImageFile[] ImageFiles
		{
			get { return _imageFiles; }
		}
		public ImageFile[] SelectedImageFiles
		{
			get
			{
				ArrayList imageFiles = new ArrayList();
				foreach (ImageFile imageFile in _imageFiles)
				{
					if (imageFile.Selected)
						imageFiles.Add(imageFile);
				}
				return (ImageFile[])imageFiles.ToArray(typeof(ImageFile));
			}
		}
		public int [] SelectedIndices
		{
			get
			{
				ArrayList indices = new ArrayList();
				for (int i = 0; i < _imageFiles.Length; i++)
				{
					if (_imageFiles[i].Selected)
						indices.Add(i);
				}
				return (int [])indices.ToArray(typeof(int));
			}
			set
			{
				foreach (ImageFile imageFile in _imageFiles)
					imageFile.Selected = false;
				foreach (int index in value)
					_imageFiles[index].Selected = true;
			}
		}
		public DirectoryInfo DirectoryInfo
		{
			get { return _directory; }
		}
		private ImageFile [] LoadFileList(DirectoryInfo directory)
		{
			FileInfo [] fis = directory.GetFiles();
			ArrayList imageFiles = new ArrayList(fis.Length);

			foreach (FileInfo fi in fis)
			{
				if ((fi.Attributes & (FileAttributes.System | FileAttributes.Hidden)) == 0)
				{
					imageFiles.Add(new ImageFile(this, fi));
				}
			}

			return (ImageFile [])imageFiles.ToArray(typeof(ImageFile));
		}
		private static ImageFile [] Merge(ImageFile [] left, ImageFile [] right)
		{
			if (right.Length == 0)
				return left;
			ArrayList imageFiles	= new ArrayList(left.Length);
			string [] leftNames		= GetNames(left);
			string [] rightNames	= GetNames(right);
			IComparer comp			= CaseInsensitiveComparer.Default;
			Array.Sort(leftNames, left, comp);
			Array.Sort(rightNames, right, comp);
			int						lndx = 0;
			int						rndx = 0;
			while ((lndx < left.Length) || (rndx < right.Length))
			{
				int rel;
				if (lndx < left.Length)
				{
					if (rndx < right.Length)
						rel = comp.Compare(leftNames[lndx], rightNames[rndx]);
					else
						rel = -1;
				}
				else
				{
					rel = +1;
				}

				if (rel == 0)
				{
					ImageFile leftFile	= left[lndx];
					ImageFile rightFile	= right[rndx];
					if (leftFile.LastWriteTime > rightFile.LastWriteTime)
					{
						imageFiles.Add(leftFile);
						rightFile.Dispose();
						rightFile = null;
					}
					else
					{
						leftFile.Dispose();
						leftFile = null;
						imageFiles.Add(rightFile);
					}
					lndx++;
					rndx++;
				}
				else if (rel < 0)
				{
					ImageFile leftFile	= left[lndx];
					imageFiles.Add(leftFile);
					lndx++;
				}
				else // (rel > 0)
				{
					ImageFile rightFile	= right[rndx];
					rightFile.Dispose();
					rightFile = null;
					rndx++;
				}
			}
			return (ImageFile [])imageFiles.ToArray(typeof(ImageFile));
		}
		public void SaveThumbsFile()
		{
			bool doSave = false;
			foreach (ImageFile imageFile in _imageFiles)
			{
				if (imageFile.LoadException == null)
				{
					doSave = true;
					break;
				}
			}
			if (doSave)
				SaveImageFiles(_thumbFilePath, _imageFiles);
		}
		private static string [] GetNames(ImageFile [] imageFiles)
		{
			string [] names = new string[imageFiles.Length];
			for (int i = 0; i < imageFiles.Length; i++)
				names[i] = imageFiles[i].Name.ToLower();
			return names;
		}
		private void SaveImageFiles(string path, ImageFile [] imageFiles)
		{
			try
			{
				FileInfo fi = new FileInfo(path);
				if (fi.Exists)
					fi.Attributes &= ~(FileAttributes.Hidden | FileAttributes.System);

				using (Stream stream = new FileStream(path, FileMode.Create, FileAccess.ReadWrite, FileShare.None))
				{
					BinaryFormatter formatter	= new BinaryFormatter();
					formatter.Serialize(stream, imageFiles);

					// store images separately to have more control over the used format
					foreach (ImageFile imageFile in imageFiles)
					{
						MemoryStream mem = new MemoryStream();
						imageFile.SaveThumb(mem, _thumbQuality);

						byte [] data = new byte[mem.Length];
						if (mem.Length > 0)
						{
							mem.Position = 0;
							mem.Read(data, 0, Convert.ToInt32(mem.Length));
						}
						formatter.Serialize(stream, data);
					}
				}
			}
			catch (Exception ex)
			{
				File.Delete(path);
				throw ex;
			}
			finally
			{
				FileInfo fi = new FileInfo(path);
				fi.Attributes |= FileAttributes.Hidden; // | FileAttributes.System;
			}
		}
		private ImageFile [] LoadImageFiles(string path)
		{
			try
			{
				ImageFile [] imageFiles;
				using (Stream stream = File.OpenRead(path))
				{
					BinaryFormatter formatter = new BinaryFormatter();
					imageFiles = (ImageFile [])formatter.Deserialize(stream);
					if (imageFiles != null)
					{
						foreach (ImageFile imageFile in imageFiles)
						{
							imageFile._folder = this;

							byte [] data = (byte [])formatter.Deserialize(stream);
							MemoryStream mem = new MemoryStream(data);
							if (mem.Length > 0)
								imageFile._thumb = Image.FromStream(mem);
						}
					}
				}
				Debug.WriteLine("Loaded '" + path + "'.");
				return imageFiles;
			}
			catch (Exception ex)
			{
				try
				{
					File.SetAttributes(path, FileAttributes.Normal);
					File.Delete(path);
				}
				catch
				{
				}
				Debug.WriteLine(ex.Message);
				return new ImageFile[0];
			}
		}

		string					_thumbFilePath;
		long					_thumbQuality;
		DirectoryInfo			_directory;
		ImageFile []			_imageFiles;
		internal HighCache		_highCache;
		ThumbLoader				_thumbLoader;
	}
}
